#include "ntapi.h"
#include <ntddk.h>
#include <WinDef.h>
#include <malloc.h>

#pragma region Globals

volatile UCHAR guz = 0;

#pragma endregion Globals

#pragma region Constants

#define ALLOCSIZE PAGE_SIZE

//#define _REAL_DELETE_

#ifdef _REAL_DELETE_
#define USE_DELETE_ON_CLOSE FILE_DELETE_ON_CLOSE
#define FILE_ACCESS FILE_GENERIC_READ|DELETE
#else
#define USE_DELETE_ON_CLOSE FILE_DIRECTORY_FILE
#define FILE_ACCESS FILE_GENERIC_READ
#endif

#define echo(x) x 
#define label(x) echo(x)__LINE__ 
#define RTL_CONSTANT_STRINGW(s) { sizeof( s ) - sizeof( (s)[0] ), sizeof( s ),(PWSTR)(s) } 
#define STATIC_UNICODE_STRING(name, str) static const WCHAR label(__)[] = L##str; static const UNICODE_STRING name = RTL_CONSTANT_STRINGW(label(__))
#define STATIC_OBJECT_ATTRIBUTES(oa, name) STATIC_UNICODE_STRING(label(m), name); static OBJECT_ATTRIBUTES oa = { sizeof oa, 0, (PUNICODE_STRING)&label(m), OBJ_CASE_INSENSITIVE }

#define RtlPointerToOffset(Base,Pointer)  ((ULONG)(((PCHAR)(Pointer))-((PCHAR)(Base))))

#pragma endregion Constants

#pragma region Structures

typedef struct _FILE_DIRECTORY_INFORMATION {
	ULONG NextEntryOffset;
	ULONG FileIndex;
	LARGE_INTEGER CreationTime;
	LARGE_INTEGER LastAccessTime;
	LARGE_INTEGER LastWriteTime;
	LARGE_INTEGER ChangeTime;
	LARGE_INTEGER EndOfFile;
	LARGE_INTEGER AllocationSize;
	ULONG FileAttributes;
	ULONG FileNameLength;
	WCHAR FileName[1];
} FILE_DIRECTORY_INFORMATION, *PFILE_DIRECTORY_INFORMATION;

typedef struct _SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX
{
	PVOID Object;
	HANDLE UniqueProcessId;
	HANDLE HandleValue;
	ULONG GrantedAccess;
	USHORT CreatorBackTraceIndex;
	USHORT ObjectTypeIndex;
	ULONG HandleAttributes;
	ULONG Reserved;
} SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX, *PSYSTEM_HANDLE_TABLE_ENTRY_INFO_EX;

typedef struct _SYSTEM_HANDLE_INFORMATION_EX
{
	ULONG_PTR NumberOfHandles;
	ULONG_PTR Reserved;
	SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX Handles[1];
} SYSTEM_HANDLE_INFORMATION_EX, *PSYSTEM_HANDLE_INFORMATION_EX;

typedef struct _KAPC_STATE
{
	LIST_ENTRY ApcListHead[2];
	PKPROCESS Process;
	UCHAR KernelApcInProgress;
	UCHAR KernelApcPending;
	UCHAR UserApcPending;
} KAPC_STATE, *PKAPC_STATE;

typedef struct _OBJECT_HANDLE_FLAG_INFORMATION {
	BOOLEAN Inherit;
	BOOLEAN ProtectFromClose;
}OBJECT_HANDLE_FLAG_INFORMATION, *POBJECT_HANDLE_FLAG_INFORMATION;

typedef struct _SYSTEM_THREAD {
	LARGE_INTEGER  KernelTime;
	LARGE_INTEGER  UserTime;
	LARGE_INTEGER  CreateTime;
	ULONG          WaitTime;
	PVOID          StartAddress;
	CLIENT_ID      ClientId;
	KPRIORITY      Priority;
	KPRIORITY      BasePriority;
	ULONG          ContextSwitchCount;
	THREAD_STATE   State;
	LONG           WaitReason;
} SYSTEM_THREAD, *PSYSTEM_THREAD;

typedef struct SYSTEM_PROCESS_INFORMATION {
	ULONG NextEntryOffset;
	ULONG NumberOfThreads;
	LARGE_INTEGER Reserved[3];
	LARGE_INTEGER CreateTime;
	LARGE_INTEGER UserTime;
	LARGE_INTEGER KernelTime;
	UNICODE_STRING ImageName;
	DWORD BasePriority;
	HANDLE ProcessId;
	HANDLE ParentProcessId;
	ULONG HandleCount;
	ULONG Reserved2[2];
	VM_COUNTERS VMCounters;
	IO_COUNTERS IOCounters;
	SYSTEM_THREAD Threads[1];
}  SYSTEM_PROCESS_INFORMATION, *PSYSTEM_PROCESS_INFORMATION;

#pragma endregion Structures

#pragma region Apis

NTKERNELAPI NTSTATUS ObSetHandleAttributes(HANDLE Handle, POBJECT_HANDLE_FLAG_INFORMATION HandleFlags, KPROCESSOR_MODE PreviousMode);
NTKERNELAPI NTSTATUS PsLookupProcessByProcessId(IN HANDLE ProcessId, OUT PEPROCESS *Process);
NTKERNELAPI VOID KeStackAttachProcess(PEPROCESS PROCESS, PKAPC_STATE ApcState);
NTKERNELAPI VOID KeUnstackDetachProcess(PKAPC_STATE ApcState);
NTKERNELAPI NTSTATUS ZwQueryDirectoryFile(IN HANDLE FileHandle, IN HANDLE Event OPTIONAL, IN PIO_APC_ROUTINE ApcRoutine OPTIONAL, IN PVOID ApcContext OPTIONAL, OUT PIO_STATUS_BLOCK IoStatusBlock, OUT PVOID FileInformation, IN ULONG Length, IN FILE_INFORMATION_CLASS FileInformationClass, IN BOOLEAN ReturnSingleEntry, IN PUNICODE_STRING FileName  OPTIONAL, IN BOOLEAN RestartScan);

#pragma endregion Apis

NTSTATUS FileHandleToUNICODE_STRING(IN	HANDLE FileHandle, OUT PUNICODE_STRING FileName) {

	PIO_STATUS_BLOCK		pioStatusBlock;
	PFILE_NAME_INFORMATION	fniFileInfo;
	NTSTATUS				status;

	pioStatusBlock = (PIO_STATUS_BLOCK)ExAllocatePoolWithTag(NonPagedPool, sizeof(IO_STATUS_BLOCK), 'TAG');

	if (pioStatusBlock == NULL)
	{
		status = STATUS_INSUFFICIENT_RESOURCES;

		DbgPrint(":Failed to allocate memory for pioStatusBlock");

		return status;
	}

	fniFileInfo = (PFILE_NAME_INFORMATION)ExAllocatePoolWithTag(NonPagedPool, sizeof(FILE_NAME_INFORMATION) + (MAX_PATH - 1), 'TAG');

	if (fniFileInfo == NULL)
	{
		status = STATUS_INSUFFICIENT_RESOURCES;

		ExFreePoolWithTag(pioStatusBlock, 'TAG');

		DbgPrint(":Failed to allocate memory for fniFileInfo");

		return status;
	}

	status = ZwQueryInformationFile(FileHandle, pioStatusBlock, fniFileInfo, (ULONG)sizeof(FILE_NAME_INFORMATION) + (MAX_PATH - 1), FileNameInformation);

	if (status != STATUS_SUCCESS)
	{
		ExFreePoolWithTag(pioStatusBlock, 'TAG');
		ExFreePoolWithTag(fniFileInfo, 'TAG');

		DbgPrint("ZwQueryInformationFile() --.STATUS %x \n", status);

		return status;
	}

	FileName->MaximumLength = fniFileInfo->FileNameLength;
	FileName->Length = fniFileInfo->FileNameLength;

	RtlCopyMemory(FileName->Buffer, &fniFileInfo->FileName, fniFileInfo->FileNameLength);

	ExFreePoolWithTag(pioStatusBlock, 'TAG');
	ExFreePoolWithTag(fniFileInfo, 'TAG');

	status = STATUS_SUCCESS;

	return status;  

}

BOOLEAN KStr_Equals(IN PUNICODE_STRING str1, IN PUNICODE_STRING str2, BOOLEAN CaseInSensitive) {

	return RtlEqualUnicodeString(str1, str2, CaseInSensitive);

}

BOOLEAN KStr_IndexOf(IN PUNICODE_STRING str, IN PUNICODE_STRING substr, OUT PUSHORT pos) {

	USHORT index;

	if (KStr_Equals(str, substr, TRUE)) {

		*pos = 0;

		return TRUE;

	}

	for (index = 0; index + (substr->Length / sizeof(WCHAR)) <= (str->Length / sizeof(WCHAR)); index++) {

		if (_wcsnicmp(&str->Buffer[index],

			substr->Buffer,

			(substr->Length / sizeof(WCHAR))) == 0) {

			*pos = index;

			return TRUE;

		}

	}

	return FALSE;

}

NTSTATUS CloseHandlesIfMatchFileName(DWORD64 id, UNICODE_STRING us)
{
	
	NTSTATUS status;
	status = STATUS_SUCCESS;
	OBJECT_HANDLE_FLAG_INFORMATION ohfi;
	ULONG ReturnLength = 1024;
	KAPC_STATE ks;
	PEPROCESS process;

	union Data {
		PVOID buf;
		PSYSTEM_HANDLE_INFORMATION_EX pshti;
	};

	union Data data;

	USHORT tmpPos = 0;

	UNICODE_STRING FName = { 0 };
	FName.Length = 1024;
	FName.MaximumLength = 1024;
	FName.Buffer = ExAllocatePoolWithTag(NonPagedPool, FName.MaximumLength, '2leN');

	if (FName.Buffer == NULL)
		return status;

	do
	{
		status = STATUS_INSUFFICIENT_RESOURCES;
		data.buf = ExAllocatePoolWithTag(PagedPool, ReturnLength, 'tset');

		if (data.buf)
		{
			if (0 <= (status = ZwQuerySystemInformation(SystemExtendedHandleInformation, data.buf, ReturnLength, &ReturnLength)))
			{
				ULONG_PTR NumberOfHandles = data.pshti->NumberOfHandles;

				if (NumberOfHandles)
				{
					SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX* Handles = data.pshti->Handles;
					do
					{

						if ((Handles->UniqueProcessId == (HANDLE)id)) {

							status = PsLookupProcessByProcessId(id, &process);

							if (NT_SUCCESS(status))
							{
								RtlZeroMemory(FName.Buffer, FName.MaximumLength);

								KeStackAttachProcess(process, &ks);

								FileHandleToUNICODE_STRING(Handles->HandleValue, &FName);

								if (KStr_IndexOf(&FName, &us, &tmpPos))
								{ 

									DbgPrint("---. FName: %wZ => Pos: %d\n", &FName, tmpPos);

									/*	ohfi.Inherit = 0;
										ohfi.ProtectFromClose = 0;
										ObSetHandleAttributes(Handles->HandleValue, &ohfi, KernelMode);
										status = ZwClose(Handles->HandleValue);
										DbgPrint("ZwClose() status: 0x%X", status);  */

								}

								KeUnstackDetachProcess(&ks);
								ObDereferenceObject(process);
							}
						}

					} while (Handles++, --NumberOfHandles);
				}
			}

			ExFreePoolWithTag(data.buf, 'tset');
		}

	} while (status == STATUS_INFO_LENGTH_MISMATCH);

	ExFreePoolWithTag(FName.Buffer, '2leN');

	return status;
}

NTSTATUS PrintProcessesUsingFile(PFILE_PROCESS_IDS_USING_FILE_INFORMATION ppiufi, UNICODE_STRING us)
{
	NTSTATUS status;

	ULONG cb = 0x8000;

	do
	{
		status = STATUS_INSUFFICIENT_RESOURCES;
		PVOID buf = ExAllocatePoolWithTag(PagedPool, cb, 'tset');

		if (buf)
		{
			if (0 <= (status = ZwQuerySystemInformation(SystemProcessInformation, buf, cb, &cb)))
			{
				union Data {
					PVOID pv;
					PBYTE pb;
					PSYSTEM_PROCESS_INFORMATION pspi;
				};

				union Data data;

				data.pv = buf;
				ULONG NextEntryOffset = 0;

				do
				{
					data.pb += NextEntryOffset;

					ULONG NumberOfProcessIdsInList = ppiufi->NumberOfProcessIdsInList;

					PULONG_PTR ProcessIdList = ppiufi->ProcessIdList;
					do
					{
						if (*ProcessIdList++ == (ULONG_PTR)data.pspi->ProcessId)
						{
							DbgPrint("%d :: %wZ\n", data.pspi->ProcessId, &data.pspi->ImageName);
							CloseHandlesIfMatchFileName(data.pspi->ProcessId, us);
							break;
						}
					} while (--NumberOfProcessIdsInList);

				} while (NextEntryOffset = data.pspi->NextEntryOffset);
			}

			ExFreePoolWithTag(buf, 'tset');
		}

	} while (status == STATUS_INFO_LENGTH_MISMATCH);

	return status;
}

NTSTATUS PrintProcessesUsingFileH(HANDLE hFile, UNICODE_STRING us)
{
	NTSTATUS status;
	IO_STATUS_BLOCK iosb;

	ULONG cb = 0, rcb = FIELD_OFFSET(FILE_PROCESS_IDS_USING_FILE_INFORMATION, ProcessIdList[64]);

	union Data {
		PVOID buf;
		PFILE_PROCESS_IDS_USING_FILE_INFORMATION ppiufi;
	};

	PVOID stack = alloca(guz);

	union Data data;

	do
	{
		if (cb < rcb)
		{
			cb = RtlPointerToOffset(data.buf = alloca(rcb - cb), stack);
		}

		if (0 <= (status = ZwQueryInformationFile(hFile, &iosb, data.ppiufi, cb, FileProcessIdsUsingFileInformation)))
		{
			if (data.ppiufi->NumberOfProcessIdsInList)
			{
				PrintProcessesUsingFile(data.ppiufi, us);
			}
		}

		rcb = FIELD_OFFSET(FILE_PROCESS_IDS_USING_FILE_INFORMATION, ProcessIdList[data.ppiufi->NumberOfProcessIdsInList]);

	} while (status == STATUS_INFO_LENGTH_MISMATCH);

	return status;
}

NTSTATUS PrintProcessesUsingFilePOA(POBJECT_ATTRIBUTES poa, UNICODE_STRING us)
{
	IO_STATUS_BLOCK iosb;
	HANDLE hFile;
	NTSTATUS status;

	DbgPrint("FileName....: %wZ \n", poa->ObjectName);

	if (0 <= (status = ZwOpenFile(&hFile, FILE_READ_ATTRIBUTES, poa, &iosb, FILE_SHARE_VALID_FLAGS, 0)))
	{
		status = PrintProcessesUsingFileH(hFile, us);

		ZwClose(hFile);
	}

	DbgPrint("Final of list handles to this process");

	return status;
}

void ntTraverse(POBJECT_ATTRIBUTES poa, ULONG FileAttributes, int nLevel, PSTR prefix)
{

	if (IoGetRemainingStackSize() < PAGE_SIZE)
	{
		DbgPrint("no stack!\n");
		return;
	}

	if (!nLevel)
	{
		DbgPrint("!nLevel\n");
		return;
	}

	NTSTATUS status;
	IO_STATUS_BLOCK iosb;
	UNICODE_STRING ObjectName;
	UNICODE_STRING MyResultString = { 0 };
	OBJECT_ATTRIBUTES myoa;
	OBJECT_ATTRIBUTES oa = { sizeof(oa), 0, &ObjectName };

	DbgPrint("DIRECTORY...: %s[<%wZ>]\n", prefix, poa->ObjectName);

#ifdef _REAL_DELETE_
	if (FileAttributes & FILE_ATTRIBUTE_READONLY)
	{
		if (0 <= ZwOpenFile(&oa.RootDirectory, FILE_WRITE_ATTRIBUTES, poa, &iosb, FILE_SHARE_VALID_FLAGS, FILE_OPEN_FOR_BACKUP_INTENT | FILE_OPEN_REPARSE_POINT))
		{
			FILE_BASIC_INFORMATION fbi = { 0 };
			fbi.FileAttributes = FILE_ATTRIBUTE_NORMAL;

			ZwSetInformationFile(oa.RootDirectory, &iosb, &fbi, sizeof(fbi), FileBasicInformation);
			ZwClose(oa.RootDirectory);
		}
	}
#endif

	if (0 <= (status = ZwOpenFile(&oa.RootDirectory, FILE_ACCESS, poa, &iosb, FILE_SHARE_VALID_FLAGS,
		FILE_SYNCHRONOUS_IO_NONALERT | FILE_OPEN_REPARSE_POINT | FILE_OPEN_FOR_BACKUP_INTENT | USE_DELETE_ON_CLOSE)))
	{
		if (FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			PVOID buffer = ExAllocatePoolWithTag(PagedPool, ALLOCSIZE, 'tset');

			if (buffer)
			{
				union Data {
					PVOID pv;
					PBYTE pb;
					PFILE_DIRECTORY_INFORMATION DirInfo;
				};

				union Data data;

				while (0 <= (status = ZwQueryDirectoryFile(oa.RootDirectory, NULL, NULL, NULL, &iosb,
					data.pv = buffer, ALLOCSIZE, FileDirectoryInformation, 0, NULL, FALSE)))
				{

					ULONG NextEntryOffset = 0;

					do
					{
						data.pb += NextEntryOffset;

						ObjectName.Buffer = data.DirInfo->FileName;

						switch (ObjectName.Length = (USHORT)data.DirInfo->FileNameLength)
						{
						case 2 * sizeof(WCHAR) :
							if (ObjectName.Buffer[1] != '.') break;
						case sizeof(WCHAR) :
							if (ObjectName.Buffer[0] == '.') continue;
						}

						ObjectName.MaximumLength = ObjectName.Length;

#ifndef _REAL_DELETE_
						if (data.DirInfo->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
#endif
						{
							ntTraverse(&oa, data.DirInfo->FileAttributes, nLevel - 1, prefix - 1);
						}
#ifndef _REAL_DELETE_
						else
#endif
						{
							//DbgPrint("%s%8I64u <%wZ>\n", prefix, DirInfo->EndOfFile.QuadPart, &ObjectName);
							PrintProcessesUsingFilePOA(&oa, ObjectName);
						}
					
					} while (NextEntryOffset = data.DirInfo->NextEntryOffset);

					if (ALLOCSIZE - iosb.Information > FIELD_OFFSET(FILE_DIRECTORY_INFORMATION, FileName[256]))
					{
						break;
					}
				}

				ExFreePoolWithTag(buffer, 'tset');

				if (status == STATUS_NO_MORE_FILES)
				{
					status = STATUS_SUCCESS;
				}
			}
		}

		ZwClose(oa.RootDirectory);
	}

	if (0 > status)
	{
		DbgPrint("---- %x %wZ\n", status, poa->ObjectName);
	}
}

void ntTraverse_()
{
	char prefix[MAXUCHAR + 1];
	memset(prefix, '\t', MAXUCHAR);
	prefix[MAXUCHAR] = 0;

	STATIC_OBJECT_ATTRIBUTES(oa, "\\??\\C:\\Program Files\\Process Hacker 2");
	ntTraverse(&oa, FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_READONLY, MAXUCHAR, prefix + MAXUCHAR);
}

void testeUnload(IN PDRIVER_OBJECT DriverObject)
{
	DbgPrint("Goodbye from teste!\n");
}

NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING  RegistryPath)
{
	DriverObject->DriverUnload = testeUnload;

	DbgPrint("Hello from teste!\n");

	ntTraverse_();

	return STATUS_SUCCESS;
}


